/*
 * ServiceStateMachine.java
 * Copyright (C) 2007 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 *
 * Created on 29. Dezember 2006, 16:38
 *
 */

package info.frantzen.testing.ssmsimulator.ssm;

import info.frantzen.testing.ssmsimulator.interpreter.javacc.DumontParser;
import info.frantzen.testing.ssmsimulator.interpreter.javacc.ParseException;
import info.frantzen.testing.ssmsimulator.interpreter.javacc.TokenMgrError;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import info.frantzen.testing.ssmsimulator.types.ComplexType;
import info.frantzen.testing.ssmsimulator.types.Enumeration;
import info.frantzen.testing.ssmsimulator.types.SimpleType;
import info.frantzen.testing.ssmsimulator.types.Type;
import info.frantzen.testing.ssmsimulator.interpreter.javacc.DumontParser.IllegalTypeException;
import java.util.List;
import java.util.Set;

/**
 * A Service State Machine (SSM) is a variant of a Symbolic Transition System (STS), dedicated
 * at specifying Web Services. It extends STS by adding support for (possibly synchronous)
 * operations.<br><br>
 * An SSM consists of:<br>
 * - a set of (data) {@link Type types}<br>
 * - a set of {@link Variable variables} (every variable is either an interaction variable or a location variable)<br>
 * - a set of {@link Operation operations}<br>
 * - a set of {@link Service services}<br>
 * - a set of {@link Message messages}<br>
 * - a set of {@link Location locations}<br>
 * - a set of {@link Switch switches}<br>
 * <br>
 * This class offers support for simulating the SSM by the following features:<br>
 * - a current location can be set<br>
 * - a current valuation of the variables can be set<br>
 * - parsers can be attached to switches to evaluate switch restrictions and execute update mappings<br>
 * <br>
 * Setting a current location and a valuation for the location variables corresponds to fixing a
 * semantical state. A semantical state is what is also represented by an {@link InstantiatedLocation instantiated location} object.
 * @author Lars Frantzen
 * @version 0.2
 */
public class ServiceStateMachine {
    
    private String wsdlURLString;
    private String namespace;
    private HashSet<Type> ssmTypes = new HashSet<Type>();
    private HashSet<Variable> vars = new HashSet<Variable>();
    private HashSet<Operation> operations = new HashSet<Operation>();
    private HashSet<Message> messages = new HashSet<Message>();
    private HashSet<Location> locations = new HashSet<Location>();
    private HashSet<Switch> switches = new HashSet<Switch>();
    private HashSet<Service> services = new HashSet<Service>();
    
    private Location currentLocation = null;
    
    /** Global symbol table for the Dumont Interpreters */
//    private Hashtable<String,Variable> symtab = null;
    
    /** Creates a new instance of ServiceStateMachine */
    public ServiceStateMachine() {
    }
    
    /**
     * Adds a data type to the SSM. A data type
     * @param t the type to be added
     */
    public void addType(Type t) {
        ssmTypes.add(t);
    }
    
    /**
     * Getter for the set of types of the SSM.
     * @return the set of types of the SSM
     */
    public HashSet<Type> getTypes() {
        return ssmTypes;
    }
    
    /**
     * Adds a variable to the SSM.
     * @param v the variable to be added.
     */
    public void addVariable(Variable v) {
        vars.add(v);
    }
    
    /**
     * Getter for the set of variables of the SSM.
     * @return the set of variables of the SSM
     */
    public HashSet<Variable> getVariables() {
        return vars;
    }
    
    /**
     * Adds an Operation to the SSM.
     * @param op the operation to be added
     */
    public void addOperation(Operation op) {
        operations.add(op);
    }
    
    /**
     * Adds a service to the SSM.
     * @param s the port to be added
     */
    public void addService(Service s) {
        services.add(s);
    }
    
    /**
     * Getter for the set of services of the SSM.
     * @return the set of ports of the SSM
     */
    public HashSet<Service> getServices() {
        return services;
    }
    
    
    /**
     * Adds a message to the SSM.
     * @param m the message to be added
     */
    public void addMessage(Message m) {
        messages.add(m);
    }
    
    public HashSet<Message> getMessages() {
        return messages;
    }
    
    /**
     * Adds a location to the SSM.
     * @param l the location to be added
     */
    public void addLocation(Location l) {
        locations.add(l);
    }
    
    public HashSet<Location> getLocations() {
        return locations;
    }
    
    /**
     * Adds a switch to the SSM.
     * @param s the switch to be added
     */
    public void addSwitch(Switch s) {
        switches.add(s);
    }
    
    /**
     * Getter for the set of swiches of the SSM.
     * @return the set of swiches of the SSM
     */
    public HashSet<Switch> getSwitches() {
        return switches;
    }
    
    /**
     * Sets the current location of the SSM. Every SSM is always in one specific current location.
     * Use this method also to set the initial location of the SSM.
     * @param loc the new current location
     */
    public void setCurrentLocation(Location loc) {
        currentLocation = loc;
    }
    
    /**
     * Getter for the current location.
     * @return the current location
     */
    public Location getCurrentLocation() {
        return currentLocation;
    }
    
    /**
     * Sets the WSDL URL of the SSM. This information is of no further use for the simulator,
     * but it may be important for external programs using the simulator.
     * @param s the string representing the WSDL URL
     */
    public void setWSDLUrlString(String s) {
        wsdlURLString = s;
    }
    
    /**
     * Getter for the WSDL URL of the SSM.
     * @return the WSDL URL
     */
    public String getWSDLUrlString() {
        return wsdlURLString;
    }
    
    /**
     * Sets the target namespace of the correspndong WSDL. This information is of no further use for the simulator,
     * but it may be important for external programs using the simulator.
     * @param s the string representing the WSDL target namespace
     */
    public void setNamespace(String s) {
        namespace = s;
    }
    
    /**
     * Getter for the namespace of the corresponding WSDL.
     * @return the namespace of the corresponding WSDL
     */
    public String getNamespace() {
        return namespace;
    }
    
    
    /**
     * Calling this method attaches parsers to the switches according to the syntax for the
     * switch restrictions and update mappings given to the switch constructor.<br>
     * <br>
     * <B>Note:</B> only consistent SSM will be processed.
     * @throws ssmsimulator.interpreter.javacc.DumontParser.IllegalTypeException The parser detected a static type inconsistency.
     * @throws ssmsimulator.ssm.ServiceStateMachine.SwitchParseException If the syntax for a switch restriction or update mapping does not conform to the
     * Dumont grammar rules, this exception is thrown.
     * @throws ssmsimulator.ssm.ServiceStateMachine.NonConsistentSSMException The SSM is not {@link #checkConsistency() consistent}.
     */
    public void attachParsersToSwitches() throws IllegalTypeException, SwitchParseException, NonConsistentSSMException {
        
        if (checkConsistency() != 0)
            throw new NonConsistentSSMException("Sorry, you can only attach parsers to a consistent SSM!");
        
//        setUpSymTab();
        
        for (Iterator it = switches.iterator(); it.hasNext() ;) {
            Switch sw = (Switch) it.next();
            
            DumontParser sr_parser = null;
            DumontParser um_parser = null;
            
            try {
                sr_parser = makeParser(sw.getSwitchRestriction(), 0, sw);
                um_parser = makeParser(sw.getUpdateMapping(), 1, sw);
            } catch (TokenMgrError tme) {
                throw new SwitchParseException("Parse error: "+sw.getSwitchRestriction() + "\n" + tme.getMessage());
            } catch (ParseException ex) {
                throw new SwitchParseException("Parse error: "+sw.getSwitchRestriction() + "\n" + ex.getMessage());
            };
            
//            DumontParser sr_parser = new DumontParser(new StringReader(sw.getSwitchRestriction()));
//            try {
//                sr_parser.SwitchRestriction();
//            } catch (TokenMgrError tme) {
//                throw new SwitchRestrictionParseException("Parse error: "+sw.getSwitchRestriction() + "\n" + tme.getMessage());
//            } catch (ParseException ex) {};
            
            sw.attachSRParser(sr_parser);
            sw.attachUMParser(um_parser);
        }
    }
    
    /**
     * Returns a Dumont parser for the passed string and kind.
     * @param s the syntax string to be parsed
     * @param kind if kind == 0 the string is parsed as a switch restriction, otherwise the string is parsed
     * as an update mapping
     * @return the Dumont parser
     * @throws ssmsimulator.interpreter.javacc.ParseException The string could not be parsed according to the Dumont grammar rules.
     * @throws ssmsimulator.interpreter.javacc.DumontParser.IllegalTypeException The parser detected a static type inconsistency.
     */
    public DumontParser makeParser(String s, int kind, Switch sw) throws ParseException, IllegalTypeException {
        DumontParser parser = new DumontParser(new StringReader(s));
        if (kind == 0)
            parser.SwitchRestriction();
        else
            parser.UpdateMapping();
        parser.setVariables(vars);
        parser.setSwitch(sw);
        parser.setSSM(this);
        parser.doTypeCheck();
        return parser;
    }
    
//    private void setUpSymTab() {
//        symtab = new Hashtable<String,Variable>();
//        for (Iterator it = vars.iterator() ; it.hasNext() ; ) {
//            Variable var = (Variable) it.next();
//            symtab.put(var.getName(), var);
//        }
//    }
    
    /**
     * Returns an instatiated location representing the current location and valuation for the
     * location variables.
     * @return the current instantiated location
     */
    public InstantiatedLocation getInstantiatedLocation() {
        Valuation currentValuation = new Valuation();
        
        for (Iterator it = vars.iterator() ; it.hasNext() ; ) {
            Variable var = (Variable) it.next();
            if (var instanceof LocationVariable)
                currentValuation.addSingleValuation(var.getName(), var.getInstance());
        }
        
        return new InstantiatedLocation(currentLocation, currentValuation);
    }
    
    /**
     * Sets the current location and valuation for the location variables according to the
     * passed instantiated location.
     * @param il the instantiated location to be set
     */
    public void setInstantiatedLocation(InstantiatedLocation il) {
        currentLocation = il.getLocation();
        setLocationVariables(il.getValuation());
    }
    
    private void setLocationVariables(Valuation val) {
        for (Iterator it = vars.iterator() ; it.hasNext() ; ) {
            Variable var = (Variable) it.next();
            if (var instanceof LocationVariable)
                var.setInstance(val.getSingleInstance(var.getName()));
        }
    }
    
    /**
     * Sets the interaction variables' valuations.
     * @param val a valuation for the interaction variables
     */
    public void setInteractionVariables(Valuation val) {
        for (Iterator it = vars.iterator() ; it.hasNext() ; ) {
            Variable var = (Variable) it.next();
            if (var instanceof InteractionVariable)
                var.setInstance(val.getSingleInstance(var.getName()));
        }
    }
    
    
    /**
     * This method is just a very first sketch to do some consistency checks on the SSM.
     * @return 0  : all right<br>
     * -1 : no current location set<br>
     * -2 : there is a non-consistent {@link Switch}<br>
     * -3 : there is a non-consistent {@link Variable}<br>
     * -4 : there is a non-consistent {@link SimpleType Simple Type} or a null type<br>
     * -5 : there is a non-consistent {@link Enumeration}<br>
     * -6 : there is a non-consistent {@link ComplexType Complex Type}<br>
     * -7 : there is a non-consistent {@link Operation}<br>
     * -8 : there is a non-consistent {@link Service} or {@link Port}<br>
     * -9 : there is a non-consistent {@link Location}<br>
     * -10: there is a non-consistent {@link Message}<br>
     * -11: there are variables having the same name<br>
     * <br>
     * See the corresponding classes for the specific meaning of "non-consistent".<br>
     * <br>
     * Other checks could be:<br>
     * - are there unreachable locations?<br>
     * - do all Types and enumerations have different names?
     */
    public int checkConsistency() {
        // check if the current location is set:
        if (currentLocation == null)
            return -1;
        
        // check the switches
        for (Iterator it = switches.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -2;
            if (!((Switch) o).checkConsistency())
                return -2;
        }
        
        // check the variables
        for (Iterator it = vars.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -3;
            if (!((Variable) o).checkConsistency())
                return -3;
        }
        
        // check the types
        for (Iterator it = ssmTypes.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -4;
            Type t = (Type) o;
            if (t instanceof SimpleType) {
                if (!((SimpleType) t).checkConsistency())
                    return -4;
            }
            if (t instanceof Enumeration) {
                if (!((Enumeration) t).checkConsistency())
                    return -5;
            }
            if (t instanceof ComplexType) {
                if (!((ComplexType) t).checkConsistency())
                    return -6;
            }
        }
        
        // check the Operations
        for (Iterator it = operations.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -7;
            if (!((Operation) o).checkConsistency())
                return -7;
        }
        
        // check the Services and Ports
        for (Iterator it = services.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -8;
            Service s = (Service) o;
            if (!(s.checkConsistency()))
                return -8;
            
            for (Iterator itt = s.getPorts().iterator(); itt.hasNext() ;) {
                o = itt.next();
                if (o == null)
                    return -8;
                Port p = (Port) o;
                if (!(p.checkConsistency()))
                    return -8;
            }
        }
        
        // check the Locations
        for (Iterator it = locations.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -9;
            if (!((Location) o).checkConsistency())
                return -9;
        }
        
        // check the Messages
        for (Iterator it = messages.iterator(); it.hasNext() ;) {
            Object o = it.next();
            if (o == null)
                return -10;
            if (!((Message) o).checkConsistency())
                return -10;
        }
        
        // check if all variable names are different.
//        HashSet<String> varNames = new HashSet<String>();
//        for (Iterator it = vars.iterator(); it.hasNext() ;) {
//            String name = ((Variable) it.next()).getName();
//            if (varNames.contains(name))
//                return -11;
//            varNames.add(name);
//        }
        
        
        return 0;
    }
    
    
    /**
     * A string representation of the SSM.
     * @return a string representation of the SSM
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        
        sb.append("Namespace: " + namespace);
        sb.append('\n');
        sb.append("WSDL: " + wsdlURLString);
        sb.append('\n');
        sb.append("Services: " + services);
        sb.append('\n');
        sb.append("Current Location: " + currentLocation);
        sb.append('\n');
        sb.append("Locations: " + locations);
        sb.append('\n');
        sb.append("Messages: " + messages);
        sb.append('\n');
        sb.append("Operations: " + operations);
        sb.append('\n');
        sb.append("Types: " + ssmTypes);
        sb.append('\n');
        sb.append("Switches: " + switches);
        sb.append('\n');
//        sb.append("symtab: " + symtab);
//        sb.append('\n');
        sb.append("Variables: " + vars);
        return sb.toString();
    }
    
    public String asDotDigraph() {
        StringBuffer sb = new StringBuffer();
        sb.append("digraph SSM {\n");
        
        Set<Location> visitedLocations = new HashSet<Location>();
        List<Location> actualLocations = new ArrayList<Location>();
        visitedLocations.add(currentLocation);
        actualLocations.add(currentLocation);
        
        boolean notAllVisited = true;
        while (notAllVisited) {
            Location loc = actualLocations.remove(0);
            actualLocations.addAll(loc.traverseDotGraph(sb));
            notAllVisited = visitedLocations.addAll(actualLocations);
        }
        
        sb.append("}\n");
        return sb.toString();
    }
    
    /**
     * This exception is thrown when a method is called which demands a consistsnt SSM, which is not the case.
     */
    public static class NonConsistentSSMException extends Exception {
        NonConsistentSSMException(String text) {
            super(text);
        }
    }
    
    /**
     * This exception is thrown when the Dumont grammar is violated.
     */
    public static class SwitchParseException extends Exception {
        SwitchParseException(String text) {
            super(text);
        }
    }
    
}



